/*
 * $Id: select.c,v 1.5 1998/05/29 20:03:22 masaki Exp $
 */

#include <stdio.h>
#include <config.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#include <New.h>
#include <unistd.h>
#include <linked_list.h>
#include <mrt.h>
#include <trace.h>
#include <select.h>
#include <version.h>
#include <errno.h>
#ifdef __linux__
#include <sys/resource.h>
#endif

#ifndef HAVE_PTHREAD_H
#include <pthread_fake.h>
#else
#include <pthread.h>
#endif /* HAVE_PTHREAD_H */

Select_Struct *SELECT_MASTER;

static void __init_select ();

/* #define MRT_DEBUG 1 */

/* init_select
 */
int 
init_select (trace_t * tr)
{
    int fd[2];

    assert (SELECT_MASTER == NULL);

    SELECT_MASTER = New (Select_Struct);
    SELECT_MASTER->ll_descriptors = LL_Create (0);
    SELECT_MASTER->trace = tr;

    pthread_mutex_init (&SELECT_MASTER->mutex_lock, NULL);

    pipe (fd);
    SELECT_MASTER->interrupt_fd[1] = fd[1];
    SELECT_MASTER->interrupt_fd[0] = fd[0];
    FD_SET (fd[0], &SELECT_MASTER->fdvar_read);
    /* setting nonblock is safer */
    fcntl (fd[1], F_SETFL, O_NONBLOCK);

    if (mrt_thread_create
	("Select Thread", NULL,
	 (thread_fn_t) __init_select, NULL) == NULL) {
	return (-1);
    }

    return (1);
}


static void 
__init_select ()
{

#ifndef HAVE_LIBPTHREAD
    return;
#else

    SELECT_MASTER->self = pthread_self ();
    init_mrt_thread_signals ();

    while (1)
	mrt_select ();
#endif /* THREAD */

}


void 
__clear_select_interrupt ()
{
    char c;

    printf ("\nhere in interrupt!\n");
    read (SELECT_MASTER->interrupt_fd[0], &c, 1);
    select_enable_fd (SELECT_MASTER->interrupt_fd[0]);

    /*exit (0); */

}


#define tmp xtmp

/* select_add_fd
 * Add a file descriptor to the list of fds which we are blocking on in a 
 * select call.
 */
int 
select_add_fd (int fd, int type_mask, void (*call_fn) (), void *arg)
{
    Descriptor_Struct *tmp;
    char c = '\0';		/* make purify happy */

    /* save us from ourselves *//* no more -- masaki */
    assert (SELECT_MASTER);

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: Adding %d\n", fd);
#endif /* MRT_DEBUG */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp) {
	if (tmp->marked_for_deletion == 1)
	    continue;
	if ((tmp->fd == fd) && (tmp->type_mask & type_mask)) {
	    trace (ERROR, SELECT_MASTER->trace,
		   "SELECT Error --same fd (%d) and mask %d:%d\n",
		   fd, tmp->type_mask, type_mask);
	    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
	    return (-1);
	}
    }
    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

    tmp = New (Descriptor_Struct);

    tmp->fd = fd;
    tmp->type_mask = type_mask;
    tmp->call_fn = call_fn;
    tmp->arg = arg;
    tmp->marked_for_deletion = 0;

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    if (type_mask & SELECT_READ)
	FD_SET (fd, &SELECT_MASTER->fdvar_read);
    if (type_mask & SELECT_WRITE)
	FD_SET (fd, &SELECT_MASTER->fdvar_write);
    if (type_mask & SELECT_EXCEPTION)
	FD_SET (fd, &SELECT_MASTER->fdvar_except);

    LL_Add (SELECT_MASTER->ll_descriptors, tmp);

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

    write (SELECT_MASTER->interrupt_fd[1], &c, 1);

    return (1);
}


/* select_delete_fd
 * Delete everything associated with this socket descriptor!
 */
int 
select_delete_fd (int fd)
{
    Descriptor_Struct *tmp;
    char c = '\0';		/* make purify happy */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: deleting %d\n", fd);
#endif /* MRT_DEBUG */

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if (tmp->fd == fd) {
	tmp->marked_for_deletion = 1;
	if (tmp->type_mask & 1)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & 2)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & 4)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_except);
    }

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
    close (fd);

    write (SELECT_MASTER->interrupt_fd[1], &c, 1);
    return (1);
}


/* select_delete_fd2
 * without closing fd
 * Delete everything associated with this socket descriptor!
 */
int 
select_delete_fd2 (int fd)
{
    Descriptor_Struct *tmp;
    char c = '\0';		/* make purify happy */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if (tmp->fd == fd) {
	tmp->marked_for_deletion = 1;
	if (tmp->type_mask & 1)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & 2)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & 4)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_except);

	pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
	write (SELECT_MASTER->interrupt_fd[1], &c, 1);
	return (1);
    }

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
#ifdef MRT_DEBUG
    /* this could happen when called from bgp */
    trace (ERROR, SELECT_MASTER->trace, "SELECT didn't find fd (%d)\n", fd);
#endif /* MRT_DEBUG */
    return (-1);
}


/* mrt_select
 */
int 
mrt_select2 (int ms)
{
    fd_set fdvar_read, fdvar_write, fdvar_except;
    Descriptor_Struct *tmp;
    struct timeval *timeout;

#ifdef HAVE_LIBPTHREAD
    timeout = NULL;
#else
    struct timeval tv;

    tv.tv_sec = ms / 1000;
    tv.tv_usec = (ms % 1000) * 1000;
    timeout = &tv;
#endif /* HAVE_LIBPTHREAD */
    fdvar_read = SELECT_MASTER->fdvar_read;
    fdvar_write = SELECT_MASTER->fdvar_write;
    fdvar_except = SELECT_MASTER->fdvar_except;

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: blocking in select\n");
#endif /* MRT_DEBUG */

    if (select (FD_SETSIZE, &fdvar_read,
		&fdvar_write, &fdvar_except, timeout) < 0) {
#ifdef MRT_DEBUG
        trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: error (%s)\n", 
	       strerror (errno));
#endif /* MRT_DEBUG */
	/*printf ("Errno %d\n", errno); exit (0); */
	return (-1);
    }
    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: processing select\n");
#endif /* MRT_DEBUG */

    /* okay, read hint to ourselves that a new socket has been added */
    if (FD_ISSET (SELECT_MASTER->interrupt_fd[0], &fdvar_read)) {
	char c;
	read (SELECT_MASTER->interrupt_fd[0], &c, 1);
    }

    tmp = LL_GetHead (SELECT_MASTER->ll_descriptors);
    while (tmp) {

	if (tmp->marked_for_deletion == 1) {
	    LINKED_LIST *next;
	    next = LL_GetNext (SELECT_MASTER->ll_descriptors, tmp);
	    LL_Remove (SELECT_MASTER->ll_descriptors, tmp);
	    Delete (tmp);
	    tmp = (Descriptor_Struct *) next;
	    continue;
	}

	if ((tmp != NULL) &&
	    (((tmp->type_mask & SELECT_READ) && FD_ISSET (tmp->fd, &fdvar_read)) ||
	     ((tmp->type_mask & SELECT_WRITE) && FD_ISSET (tmp->fd, &fdvar_write)) ||
	     ((tmp->type_mask & SELECT_EXCEPTION) && FD_ISSET (tmp->fd, &fdvar_except)))) {

/* Clearing the bit is required because sometimes this routine is called 
   again before reading data waiting. */
	    if (tmp->type_mask & SELECT_READ)
		FD_CLR (tmp->fd, &SELECT_MASTER->fdvar_read);
	    if (tmp->type_mask & SELECT_WRITE)
		FD_CLR (tmp->fd, &SELECT_MASTER->fdvar_write);
	    if (tmp->type_mask & SELECT_EXCEPTION)
		FD_CLR (tmp->fd, &SELECT_MASTER->fdvar_except);

#ifdef MRT_DEBUG
	    trace (TRACE, SELECT_MASTER->trace,
		   "mrt_select (%d,%x,%x,%x) calling %x(%x)\n", tmp->fd,
		   *(int *) &fdvar_read, *(int *) &fdvar_write, *(int *) &fdvar_except,
		   tmp->call_fn, tmp->arg);
	    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: About to call_fd\n");
#endif /* MRT_DEBUG */

	    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
	    tmp->call_fn (tmp->arg);
#ifdef MRT_DEBUG
	    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: done\n");
#endif /* MRT_DEBUG */
	    return (1);
	}

	tmp = LL_GetNext (SELECT_MASTER->ll_descriptors, tmp);
    }

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
    return (0);
}


int mrt_select () {
  return (mrt_select2 (1000));
}


/* there is no way to disable for write only */

int 
select_disable_fd (int fd)
{
    Descriptor_Struct *tmp;
    char c = '\0';		/* make purify happy */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if (tmp->fd == fd) {
	if (tmp->type_mask & SELECT_READ)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & SELECT_WRITE)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & SELECT_EXCEPTION)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_except);

	pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

	write (SELECT_MASTER->interrupt_fd[1], &c, 1);

	return (1);
    }

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
#ifdef MRT_DEBUG
    trace (FATAL, SELECT_MASTER->trace, "SELECT doesn't find fd (%d)\n", fd);
#endif	/* MRT_DEBUG */
    return (-1);
}


/* given a mask (SELECT_READ| SELECT_WRITE_ | SELECT_EXCEPTION),
 * disable a socket that matches the fd AND the mask
 */
int 
select_disable_fd_mask (int fd, int mask)
{
    Descriptor_Struct *tmp;
    char c = '\0';		/* make purify happy */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if ((tmp->fd == fd) && (tmp->type_mask & mask)) {
	if (tmp->type_mask & SELECT_READ)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & SELECT_WRITE)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & SELECT_EXCEPTION)
	    FD_CLR (fd, &SELECT_MASTER->fdvar_except);

	pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

	write (SELECT_MASTER->interrupt_fd[1], &c, 1);

	return (1);
    }

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
#ifdef MRT_DEBUG
    trace (FATAL, SELECT_MASTER->trace, "SELECT doesn't find fd (%d)\n", fd);
#endif	/* MRT_DEBUG */
    return (-1);
}



/* select_enable_fd
 * the mrt_select master must be notified that a socket is ready for selection again
 * after the select call routine has fired
 */
int 
select_enable_fd (int fd)
{
    Descriptor_Struct *tmp = NULL;
    char c = '\0';		/* make purify happy */

    if (fd < 0) {
	return (-1);
    }

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: select_enable_fd %d\n", fd);
#endif	/* MRT_DEBUG */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if (tmp->fd == fd) {
	if (tmp->type_mask & 1)
	    FD_SET (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & 2)
	    FD_SET (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & 4)
	    FD_SET (fd, &SELECT_MASTER->fdvar_except);

	pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

	write (SELECT_MASTER->interrupt_fd[1], &c, 1);
	return (1);
    }

    /* this fd does not exist! */
    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
#ifdef MRT_DEBUG
    trace (FATAL, SELECT_MASTER->trace, "SELECT doesn't find fd (%d)\n", fd);
#endif
    return (-1);
}


int 
select_enable_fd_mask (int fd, int mask)
{
    Descriptor_Struct *tmp = NULL;
    char c = '\0';		/* make purify happy */

    if (fd < 0) {
	return (-1);
    }

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: select_enable_fd %d\n", fd);
#endif	/* MRT_DEBUG */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp)
	if ((tmp->fd == fd) && (tmp->type_mask & mask)) {
	if (tmp->type_mask & SELECT_READ)
	    FD_SET (fd, &SELECT_MASTER->fdvar_read);
	if (tmp->type_mask & SELECT_WRITE)
	    FD_SET (fd, &SELECT_MASTER->fdvar_write);
	if (tmp->type_mask & SELECT_EXCEPTION)
	    FD_SET (fd, &SELECT_MASTER->fdvar_except);

	pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

	write (SELECT_MASTER->interrupt_fd[1], &c, 1);
	return (1);
    }

    /* this fd does not exist! */
    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
#ifdef MRT_DEBUG
    trace (FATAL, SELECT_MASTER->trace, "SELECT doesn't find fd (%d)\n", fd);
#endif
    return (-1);
}

/* called on exit */
int
select_shutdown ()
{
    Descriptor_Struct *tmp;

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp) {
	printf ("Shutting down %d\n", tmp->fd);
	/*if (shutdown (tmp->fd, 2) != 0) {
	   perror ("error in shutdown");
	   }
	   t_snddis (tmp->fd, NULL); */
	if (close (tmp->fd) != 0)
	    perror ("error in close");
    }

    return (1);
}


/* select_add_fd
 * Add a file descriptor to the list of fds which we are blocking on in a 
 * select call.
 */
int 
select_add_fd_off (int fd, int type_mask, void (*call_fn) (), void *arg)
{
    Descriptor_Struct *tmp;

    /* save us from ourselves *//* no more -- masaki */
    assert (SELECT_MASTER);

#ifdef MRT_DEBUG
    trace (TR_TRACE, SELECT_MASTER->trace, "SELECT: Adding %d\n", fd);
#endif /* MRT_DEBUG */

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    LL_Iterate (SELECT_MASTER->ll_descriptors, tmp) {
	if (tmp->marked_for_deletion == 1)
	    continue;
	if ((tmp->fd == fd) && (tmp->type_mask & type_mask)) {
	    trace (ERROR, SELECT_MASTER->trace,
		   "SELECT Error --same fd (%d) and mask %d:%d\n",
		   fd, tmp->type_mask, type_mask);
	    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);
	    return (-1);
	}
    }
    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

    tmp = New (Descriptor_Struct);

    tmp->fd = fd;
    tmp->type_mask = type_mask;
    tmp->call_fn = call_fn;
    tmp->arg = arg;
    tmp->marked_for_deletion = 0;

    pthread_mutex_lock (&SELECT_MASTER->mutex_lock);

    /*if (type_mask & SELECT_READ) FD_SET (fd, &SELECT_MASTER->fdvar_read);
       if (type_mask & SELECT_WRITE) FD_SET (fd, &SELECT_MASTER->fdvar_write);
       if (type_mask & SELECT_EXCEPTION) FD_SET (fd, &SELECT_MASTER->fdvar_except); */

    LL_Add (SELECT_MASTER->ll_descriptors, tmp);

    pthread_mutex_unlock (&SELECT_MASTER->mutex_lock);

    /*write (SELECT_MASTER->interrupt_fd[1], &c, 1); */

    return (1);
}
