/* 
 * $Id: user_util.c,v 1.13 1998/09/01 15:14:25 masaki Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <mrt.h>
#include <trace.h>
#include <select.h>
#include <stdarg.h>
#include <user.h>


static void add_last_token (char *ctoken, char *last_ctoken, char *last_utoken,
			    LINKED_LIST *ll_last);

/* uii_tokenize
 * tokenize a command string (e.g. "show ip access")
 */
LINKED_LIST *uii_tokenize (char *buffer, int len) {
  char command[MAXLINE], *last, *utoken;
  LINKED_LIST *ll;

  /* I want to use Delete in place of free, but it's a macro. */
  ll = LL_Create (LL_DestroyFunction, free, 0);

  /* tokenize uii string */
  memset (command, 0, MAXLINE);
  memcpy (command, buffer, len);

  utoken = strtok_r (command, " ", &last);
  while (utoken != NULL) {
    LL_Add (ll, strdup (utoken));
    utoken = strtok_r (NULL, " ", &last);
  }

  return (ll);
}



/* uii_tokenize
 * tokenize a command string (e.g. "show ip access")
 */
LINKED_LIST *uii_tokenize_choices (char *buffer, int len) {
  char command[MAXLINE], *last, *utoken, *cp;
  LINKED_LIST *ll;

  ll = LL_Create (LL_DestroyFunction, free, 0);

  /* tokenize uii string */
  memset (command, 0, MAXLINE);
  memcpy (command, buffer, len);

  cp = command +1;
  cp[strlen(command) -1] = '\0';

  utoken = strtok_r (cp, "|", &last);
  while (utoken != NULL) {
    LL_Add (ll, strdup (utoken));
    utoken = strtok_r (NULL, "|", &last);
  }

  return (ll);
}


/* uii_token_match
 */
int uii_token_match (char *ctoken, char *user_token) {
  char *cp, *buf, *last;
  int okay;

  /* match one in the set */
  if (ctoken[0] == '(') {
    cp = buf = strdup ((ctoken +1));
    cp[strlen(cp)-1] = '\0'; /* get rid of terminating ) */
    cp = strtok_r (cp, "|", &last);
    
    okay = 0;
    while (cp != NULL) {
      if (uii_token_match (cp, user_token)) {
	okay = 1;
	break;
      }
      cp = strtok_r (NULL, "|", &last);
    }

    Delete (buf);
    return (okay);
  }


  if (ctoken[0] == '%') {
    int i;

    switch (ctoken[1]) {
    case 'p': /* ipv4 prefix */
      if (is_ipv4_prefix (user_token))
	return (1);
      break;
#ifdef HAVE_IPV6
    case 'P': 
      if (is_ipv6_prefix (user_token)) 
	return (1);
      break;
#endif /* HAVE_IPV6 */
    case 'a': /* ipv4 prefix */
      if (is_ipv4_prefix (user_token) && !strchr (user_token, '/'))
	return (1);
      break;
#ifdef HAVE_IPV6
    case 'A': 
      if (is_ipv6_prefix (user_token) && !strchr (user_token, '/')) 
	return (1);
      break;
#endif /* HAVE_IPV6 */
    case 'M': 
      if (strchr (user_token, '/'))
	return (0);
    case 'm': 
      if (is_ipv4_prefix (user_token))
	return (1);
#ifdef HAVE_IPV6
      if (is_ipv6_prefix (user_token)) 
	return (1);
#endif /* HAVE_IPV6 */
      break;
    case 'n': /* name */
      if (!isalpha (user_token[0]))
	return (0);
      for (i = 1; user_token[i]; i++) {
        if (!isalnum (user_token[i]))
	    return (0);
      }
      return (1);
      break;
    case 's': /* any string */
      return (1);
      break;
    case 'd': /* integer */
      for (i = 0; user_token[i]; i++) {
        if (!isdigit (user_token[i]))
	    return (0);
      }
      return (1);
      break;
    default:
      return (0);
      break;
    }
    return (0);
  }

  /* a literal */
  if (!strncasecmp (ctoken, user_token, strlen (user_token))) 
    return (1);

  return (0);
}


/* find_matching_commands
 */
LINKED_LIST *find_matching_commands (int state, LINKED_LIST *ll_tokens,
				     LINKED_LIST *ll_last) {
  char *utoken, *ctoken, *choice;
  LINKED_LIST *ll_match, *ll_choices;
  uii_command_t *uii_command;
  char *last_ctoken= NULL, *last_utoken = NULL;
  int skip = 0;
  int optional;

  ll_match = LL_Create (0);

  /* okay, now run through the list of commands and see who matches most */
  LL_Iterate (UII->ll_uii_commands, uii_command) {
    if (state >= UII_CONFIG && uii_command->state == UII_CONFIG_ALL) {
	/* OK */
    }
    else if (state != uii_command->state)
      continue;

    last_ctoken = ctoken = LL_GetHead (uii_command->ll_tokens);
    last_utoken = utoken = LL_GetHead (ll_tokens);
    skip = 0;

    /* for password and comments -- see if any of the first part of command matches */
    if ((utoken != NULL) && (uii_command->flag & COMMAND_MATCH_FIRST)) {
      if (!strncasecmp (uii_command->string, utoken, strlen (uii_command->string))) {
	 LL_Add (ll_match, uii_command);
	 continue;
      }
    }

    /* command may be prefaced with a no -- special case */
    /* if ((utoken != NULL) && (!strcasecmp ("no", utoken)))*/
    /* last_utoken = utoken = LL_GetNext (ll_tokens, utoken);*/

    optional = 0;
    while ((utoken != NULL) && (ctoken != NULL)) {

      /* accept everything else to end of line*/
      if (!strcmp (ctoken, "%S"))
	break;
      
      /* optional command */
      if (ctoken[0] == '[') {
	if (!uii_token_match ((ctoken+1), utoken)) {
	  skip = 1;
	  break;
	}
      }
      /* mandatory command */
      else if (!uii_token_match (ctoken, utoken)) {
	skip = 1;
	break;
      }

      last_ctoken = ctoken;
      last_utoken = utoken;
      ctoken = LL_GetNext (uii_command->ll_tokens, ctoken);
      utoken = LL_GetNext (ll_tokens, utoken);
    }

    if (skip == 1) continue;
    if ((utoken != NULL) && (ctoken == NULL)) continue;

    if (ll_last == NULL) {
      LL_Add (ll_match, uii_command);
      continue;
    }

    /* we need to fill in last matched token (for tab completion and help) */
    if ((ctoken != NULL) && ((ctoken[0] == '(') || (ctoken[0] == '['))) {
      ll_choices = uii_tokenize_choices (ctoken, strlen (ctoken));
      LL_Iterate (ll_choices, choice) {
	add_last_token (choice, last_ctoken, last_utoken, ll_last);
	LL_Add (ll_match, uii_command);
      }
    }
    else if ((last_ctoken != NULL) && 
	     ((last_ctoken[0] == '(') || (last_ctoken[0] == '['))) {
      ll_choices = uii_tokenize_choices (last_ctoken, strlen (last_ctoken));
      LL_Iterate (ll_choices, choice) {
	add_last_token (ctoken, choice, last_utoken, ll_last);
	LL_Add (ll_match, uii_command);
      }
    }
    else {
      add_last_token (ctoken, last_ctoken, last_utoken, ll_last);
      LL_Add (ll_match, uii_command);
    }
  }

  return (ll_match);
}




static void add_last_token (char *ctoken, char *last_ctoken, char *last_utoken,
			    LINKED_LIST *ll_last) {

  /* done with current command -- get next */
  if ((last_utoken == NULL) ||
      ((last_ctoken != NULL) && 
       ((!strcmp (last_ctoken, last_utoken)) || (last_ctoken[0] == '%')))) {
    if (ctoken != NULL)
      LL_Add (ll_last, strdup (ctoken));
    else
      LL_Add (ll_last, strdup ("<cr>"));
    return;
  }

  if (last_ctoken == NULL) {
    LL_Add (ll_last, strdup ("<cr>"));
    return;
  }
      
  LL_Add (ll_last, strdup (last_ctoken));
  return;
}




/* uii_send_data
 * send formatted data out a socket
 */
int uii_send_data (uii_connection_t * uii_connection, ...) {
    va_list ap;
    char *format, *newline;
    char line[MAXLINE];
    char *cp = line;

    if (uii_connection == NULL)
	return (-1);
    if (uii_connection->sockfd_out < 0 && uii_connection->sockfd < 0)
	return (-1);

    va_start (ap, uii_connection);
    format = va_arg (ap, char *);
    vsprintf (cp, format, ap);

    while ((newline = strchr (cp, '\n'))) {
        int len = newline - cp + 1;
	if (len >= 2 && cp[len-2] == '\r') {
            if (uii_connection->sockfd_out >= 0) {
	        if (write (uii_connection->sockfd_out, cp, len-2)  < len-2 ||
		    write (uii_connection->sockfd_out, "\n", 1) < 1) {
		    uii_connection->sockfd_out = -1; /* turn off redirection */
		    return (-1);
	        }
            }
            else if (uii_connection->sockfd >= 0) {
	        if (write (uii_connection->sockfd, cp, len) < len) {
		    uii_destroy_connection (uii_connection);
		    return (-1);
	        }
    	    }
	}
	else {
    	    if (uii_connection->sockfd_out >= 0) {
	        if (write (uii_connection->sockfd_out, cp, len) < len) {
		    uii_connection->sockfd_out = -1; /* turn off redirection */
		    return (-1);
	        }
            }
            else if (uii_connection->sockfd >= 0) {
	        if (write (uii_connection->sockfd, cp, len-1) < len-1 ||
	            write (uii_connection->sockfd, "\r\n", 2) < 2) {
		    uii_destroy_connection (uii_connection);
		    return (-1);
	        }
    	    }
	}
	cp = newline + 1;
    }
    if (uii_connection->sockfd_out >= 0) {
	int len = strlen (cp);
        if (write (uii_connection->sockfd_out, cp, len) < len) {
	    uii_connection->sockfd_out = -1; /* turn off redirection */
	    return (-1);
	}
    }
    else if (uii_connection->sockfd >= 0) {
	int len = strlen (cp);
        if (write (uii_connection->sockfd, cp, len) < len) {
	    uii_destroy_connection (uii_connection);
	    return (-1);
        }
    }
    return (1);
}
